// Lecture 11, Smart pointers example, Dragana Laketic

#include "std_lib_facilities.h"
#include "houston.h"

// We can return unique_ptr variable from a function.
unique_ptr<Sattelite> createUniqueSattelite(string satName) {
    std::unique_ptr<Sattelite> ptr = std::make_unique<Sattelite>();
    ptr->setName(satName);

    return ptr;
}

// Call by reference!
// When we pass unique pointer as a parameter to a function,
// it is best practice to do it by reference. In that way we
// avoid calling std::move() function on the unique pointer
// where the function is called in the program.
void printSatteliteUnique(std::unique_ptr<Sattelite>& peker) {
    cout << "Funksjon: " << peker->getName() << endl;
}

int main() {
        // An example of a declaration of unique pointer:
        // ptrSat is a unique pointer to a variable of type Sattelite
        // which is created by a Sattelite default constructor.
        // We use make_unique<Sattelite>() function to create a unique pointer
        // to an object of type Sattelite which is created by Sattelite's
        // default constructor since there is an empty parameter list.
        // For:
        // std::unique_ptr<Sattelite> ptrSat = make_unique<Sattelite>("Starlink");
        // a constructor which takes one string parameter would be called and
        // a unique pointer to that object would be created.
        // We can also initialise a unique pointer without calling make_unique<>():
        // std::unique_ptr<Sattelite> yetPtr{new Sattelite{"Rosetta"}};
    std::unique_ptr<Sattelite> ptrSat = make_unique<Sattelite>();
        // Unique pointers cannot be copied - they are unique!
        // We can use function move to transfer the ownership of
        // an object to another unique pointer like in the example below:
    std::unique_ptr<Sattelite> ptrSat1 = std::move(ptrSat); // When this line is run,
                                                        // ptrSat will become invalid
                                                        // and generate an error if we
                                                        // attempt to perform any operation
                                                        // on it.

        // shared pointers are declared in the same way as unique pointers:
    std::shared_ptr<Sattelite> sharedPtr =
                                std::make_shared<Sattelite>();
    std::shared_ptr<Sattelite> anotherSharedPtr =
                                std::make_shared<Sattelite>("Apollo");
    std::shared_ptr<Sattelite> yetAnotherSharedPtr{new Sattelite{"Rosetta"}};
    std::shared_ptr<Sattelite> copyShared;
    std::shared_ptr<Sattelite> retShared;
    Sattelite sat2{"AlphaFold"}; // used in the examples below


        // We can use unique pointers just like raw pointers we have worked
        // with until now.
    cout << "My sattelite name: " << ptrSat1->getName() << endl;
    ptrSat1->setName("Betha");
    cout << "My sattelite name: " << ptrSat1->getName() << endl;

        // When passing a unique pointer as parameter to a function,
        // the best practice is to pass it by reference and in that
        // way avoid calling move() function - as for the example of
        // printSatteliteUnique() - see the definition of this function
        // at the top of the file:
    printSatteliteUnique(ptrSat1);

    // shared_ptr - some examples (although we have not had time to go
    // through all examples in the class)
    // Shared pointers are another type of smart pointers. They can be used
    // to get access to the object pointed to in hte same way as all pointers:
    cout << "Smart pointer points to " << sharedPtr->getName() << endl;
    cout << "Another smart pointer points to " << anotherSharedPtr->getName() << endl;
    cout << "Yet another smart pointer points to " << yetAnotherSharedPtr->getName() << endl;

        // Unlike unique pointers, shared pointers can be copied:
    copyShared = sharedPtr;
    cout << "Copied shared pointer points to " << copyShared->getName() << endl;
    retShared = sharedPtr;
    cout << "Second sopy points to " << retShared->getName() << endl;
        // Method use_count() can tell us how mamy references there are to the same object:
    cout << "Total num of shared pointer references " << copyShared.use_count() << endl;
    copyShared = nullptr; // This operation will decrease a reference count:
    cout << "Total num of shared pointer references a: " << sharedPtr.use_count() << endl;
    retShared = nullptr; // ... and once more
    cout << "Total num of shared pointer references b: " << sharedPtr.use_count() << endl;

    // Neither unique nor shared pointer require us to explicitly release allocated memory!!!
    // That is why we call them smart.

    return 0;
}
