// Lecture 12, Example 2: Exception-handling with try-catch-throw, Dragana Laketic

#include "std_lib_facilities.h"
#include <cassert>

    // An example class which is used to handle exceptions.
    // User-defined classes for exception handling can be
    // designed according to the needs of our program. For more
    // complex programs they can be implemented as class hierarchies.
    // In our case, it is sufficient to provide an opportunity for
    // a user to pass the message on hte exception that happened.
    // What is thrown as an exception in the program is an instance
    // of this class.
class MyException {
    string message;
public:
    MyException() : message("MyException happened!") {
        cout << message << endl;
    }
    MyException(string strMessage) : name(strMessage) {
        cout << message << endl;
    }
};

    // An example of a class which is just used for instantiating
    // a dynamically allocated object in the example of exception safety.
    // The class is known from previous lectures. What is important to
    // note here is that we have output to a terminal each time a constructor
    // or a destructor is called. it will help us follow how memory is handled -
    // allocated and freed during hte program run.
class Sattelite {
    string name;
public:
    Sattelite() : name{"Starlink"} {
        cout << "Constructing sattelite!" << endl;
    }
    Sattelite(string strName) : name{strName} {}
    ~Sattelite() {
        cout << "Destructing sattelite, freeing memory!" << endl;
    }
};

    // Definition of a function which is used in the example
    // of exception safety.
void badPractice() {
    Sattelite* ptrSat = new Sattelite();
    throw runtime_error("Bad practice at runtime!");
        // Mark that this line of code we shall never reach.
        // Memory leakage happens because of the dynamically
        // allocated memory for the object Sattelite is never
        // called to be released by the operating system.
    delete ptrSat;
}

    // Definition of a function which is used in the example
    // of exception safety.
void goodPractice() {
    shared_ptr<Sattelite> ptrSat = make_shared<Sattelite>();
    throw runtime_error("Good practice at runtime!");
    // Mark that befoore we leave the scope of this function upon
    // the exception being thrown, the memory allocated for a Sattelite
    // object pointed to by shared pointer ptrSat is released and freed
    // by shared pointer (there is a destructor output in the terminal
    // when you run the example!). This illustrates one of the recommendations
    // for writing exception safe code.
}

int main() {
        // Declaration of three variables which are used in
        // the example for exception handling. The first one is
        // considered a given value as initialised at the declaration time.
        // The second one is read from the terminal and provided by the user.
        // The third one is calculated based on the first two.
    int numberOfTents{20};
    int numberOfScouts;
    int numOfScoutsPerTent;

    cout << "Input number of scouts:" << endl;
    cin >> numberOfScouts;

        // Example of a try-catch block.
        // (Run this bit of code with various input values)
        // Within the try-block we write the code for which we want
        // to make sure that exceptional situations are handled safely.
        // For our case we inspect the values which are provided by
        // the user and dependent on these values we throw various types
        // of exceptions. Note how exceptions can be user-defined - as for
        // the example used for the instance of class MyException - and exceptions
        // from the C++ STL standard library.
        // Note also that we throw exceptions by value!
    try {
        if (numberOfScouts < 0) {
                // Here we throw a user-defined exception. In this case.
                // it is an instance of the class MyException.
                // NOTE: throw by value"
            throw MyException("There cannot be negative number of scouts!");
        }
        if (numberOfScouts % 5 != 0) {
                // In this case we throw a user-defined exception as well.
            throw MyCustomException("Scouts must be in groups of 5!");
        }
        if (numberOfScouts == 0) {
                // Here we throw a runtime_error exception from the C++ STL
                // standard library. See headers <exception> and <stdexcept>
                // for more details on these exceptions.
            throw std::runtime_error("There must be some scouts present at the gathering!");
        }
        if (numberOfScouts < 20) {
                // Yet another example of throwing a runtime_error exception
                // from standard library
            throw std::runtime_error("Too few scouts to attend the gathering!");
        }
            // The code below will be run if none of the exceptions above have been thrown.
        numOfScoutsPerTent = numberOfScouts / numOfTents;
    } catch (std::exception& e) { // NOTE: catch by reference!
            // This is the first catch-block for the try-block above.
            // Mark how we can have more than one catch-block per one
            // try-block. That is because more than one type of an exception
            // may occur during the code run and we can provide a catch-block
            // per type of an exception. In this catch-block we handle exceptions
            // from C++ standard library (see <exception> and <stdexcept>). Note
            // how we use what() method of the exception class to obtain more
            // information about the exception that occurred.
        cout << e.what() << endl;
    } catch (MyException& e) {
            // In this catch-block we handle exceptions of the user-defined type
            // class MyException
        cout << "Catching user-defined exception!" << endl;
    } catch(...) {
            // catch-block (...) handles any other exception type,
            // i.e., if an exception was thrown of a type not listed
            // in previous catch-statements, then this block is run.
        cout << "Unknown exception happened!" << endl;
    }

        // Here is an example of the code which does not (in the first case)
        // and which does (in the second case) follow the recommendations
        // for exception safety.
        // In both cases the exception is handled but in the first case
        // the destructor of the dynamically allocated object Sattelite
        // is not called because the code never ran the delete statement
        // at the end of the badPractice() function.
    cout << "Bad Practice Call:" << endl;
    try {
        badPractice();
    } catch (std::exception& e) {
        cout << e.what() << endl;
    } catch (...) {
        cout << "Unknown bad practice" << endl;
    }
    cout << "-----------------------------" << endl;
    cout << "Good Practice Call:" << endl;
        // In this case, since we used smart pointer (shared pointer)
        // to a dynamically allocated object, the destructor of that
        // object was called when an exception in the function goodPractice()
        // was thrown and the shared pointer went out of the scope where
        // it was defined, i.e., the scpe of that function. Unlike the case above
        // the memory held by operating system for Sattelite object is released.
        // Look at the messages in the terminal when you run the program.
        // Can you see where the memory leakage happened?
    try {
        goodPractice();
    } catch (std::exception& e) {
        cout << e.what() << endl;
    }  catch (...) {
        cout << "Unknown bad practice" << endl;
    }

    return 0;
}