// Lecture 13, Examples 3 and 4: STL containers and algorithms, Dragana Laketic

#include "std_lib_facilities.h"

#include <queue>
#include <stack>
#include <list>
#include <algorithm>
#include <cassert>

int main() {
    // STL containers
    std::queue<int> queue; // Queue is a First-In-First-Out (FIFO) data
                        // structure. In STL C++ standard library it is
                        // implemented as a template.
    queue.push(1); // We use push() method to enqueue an element in queue
    queue.push(2);
    queue.push(3);

    while(!queue.empty()) { // The value of an element of a queue which is
                            // first to leave the queue when we call pop() method
                            // can be accessed by front() method.
        cout << queue.front() << endl;
        queue.pop();
    }
    assert(queue.size() == 0); // repetition from last lecture: use of assert
                            // to make sure all the elements of the queue have left
                            // the queue (have been popped from the queue).

    std::stack<string> stack; // Stack is a Last-In-First-Out data structure. That means
                        // that the last element placed on stack will be the first to
                        // leave the stack when we call stack's pop() method.
    stack.push("First on Stack");
    stack.push("Second on Stack");
    stack.push("Third on Stack");

    while(!stack.empty()) {
        cout << stack.top() << endl; // We use top() method to get the values of
                                    // the element which is placed last on stack.
                                    // It will be the first element to leave the stack
                                    // when we call pop() method.
        stack.pop();
    }
    // Note how we use push() and pop() methods for two different containers -
    // a queue and a stack - but how these methods work in accordance to the
    // definition of each of them, i.e., the definition of a FIFO and a definition
    // of a LIFO data structure.


        // Back to vector. Although we know it very well, now we understand
        // a "bigger picture" as well: it is yet another template container
        // class from STL library.
    std::vector<int> vec{50, 70, 20, 40, 80, 10, 90};
        // Usually we traverse its elements in a for-loop like this:
    for (int i = 0; i < vec.size(); i++) {
        cout << vec.at(i) << endl; // Here we use a range-checked method at()
                                // to access the vector's elements.
    }
    cout << endl;

        // Here is an example of traversing the same vector with the use
        // of iterators. Iterators are members of STL containers which contain
        // a pointer to an element of the container and thereby enable us to
        // access the values stored in the container's elements.
        // Note that STL containers also provide begin() and end() methods
        // which return an iterator which points to the first element of
        // the container and to the location after the last element of the
        // container. In other words, an STL container contains the elements in
        // the interval [STLcontainer.begin(), STLcontainer.end()) - where
        // mathematical notation for open and close limit of the interval applies.
    cout << "Vector traversed with iterator: " << endl;
        // Pay attention to how the iterator type is declared! It reflects
        // the instantiated template of a container!
    for(std::vector<int>::iterator it = vec.begin(); it != vec.end(); it++) {
        cout << *it << endl;
    }
    cout << endl;

    std::list<int> lst{1, 2, 3}; // Linked lists are containers which use pointers to connect
                        // the elements into it. There are two types of lists:
                        // singly linked lists and doubly linked lists. An element
                        // of a singly linked list beside the data value contains also
                        // a pointer to the next element in the list, while the doubly
                        // linked lists elements contain a pointer to a next element in
                        // the list and a pointer to the previous element in the list.
                        // With each insersion of a new element, search for an element of
                        // certain value or erasure of an element, the pointers to some
                        // memory locations where individual elements are stored must also
                        // be handled. This all takes much of the computing resources and
                        // increases the time needed for processing. Therefore, unless really
                        // necesary within the application at hand (or the task at the exam),
                        // they are advised to be avoided. Use vectors instead and
                        // whenever you can!

    cout << "List with iterator: " << endl;
        // Pay attention to how the type of the iterator is declared in this case!
        // Compare it with the iterator type from the vector traversal example!
    for (std::list<int>::iterator it = lst.begin(); it != lst.end(); it++) {
        cout << *it << endl;
    }
    cout << endl;

        // There exist also a reverse iterator which is analogous to an iterator
        // discussed above with the distinction in the direction of the traversal
        // which is illustrated in the following example:
    cout << "Reverse iterator:" << endl;
    for (std::list<int>::reverse_iterator rit = lst.rbegin(); rit != lst.rend(); rit++) {
        cout << *rit << endl;
    }
    cout << endl;

    // STL algorithms

    // Iterators are especially handy when it comes to using algorithms from
    // the STL standard library.

        // An example of calling the sort() algorithm for the first 4
        // elements of the vector vec defined above:
    std::sort(vec.begin(), vec.begin() + 4);
    cout << "Partially sorted vector:" << endl;
    for(std::vector<int>::iterator it = vec.begin(); it != vec.end(); it++) {
        cout << *it << endl;
    }
    cout << endl;

        // An example of using an algorithm from the STL library which
        // processes the elements of two different types of containers.
        // We provide iterators to the copy algorithm and invoke the copying
        // of list lst elements from the first to the last to the first element
        // of the vector and the subsequent ones:
    std::copy(lst.begin(), lst.end(), vec.begin());
    cout << "Vector after copying from the list:" << endl;
    for(std::vector<int>::iterator it = vec.begin(); it != vec.end(); it++) {
        cout << *it << endl;
    }
    cout << endl;
        // Note that the checks of the container limits are not performed
        // so you need to take care of it. Test what would happen if instead
        // of a vec.begin() you provided vec.end() / vec.end()-1 / vec.end()-2.

        // Am example of fill algorithm:
    std::fill(vec.begin(), vec.begin()+5, 500);
    cout << "Vector after fill:" << endl;
    for(std::vector<int>::iterator it = vec.begin(); it != vec.end(); it++) {
        cout << *it << endl;
    }
    cout << endl;

    return 0;
}
